/*
 * Project: MoleCuilder
 * Description: creates and alters molecular systems
 * Copyright (C)  2014 Frederik Heber. All rights reserved.
 * 
 *
 *   This file is part of MoleCuilder.
 *
 *    MoleCuilder is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    MoleCuilder is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with MoleCuilder.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * FitCompoundPotentialAction.cpp
 *
 *  Created on: Sep 10, 2014
 *      Author: heber
 */

// include config.h
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

// needs to come before MemDebug due to placement new
#include <boost/archive/text_iarchive.hpp>

#include "CodePatterns/MemDebug.hpp"

#include <algorithm>
#include <boost/filesystem.hpp>
#include <fstream>

#include "Actions/PotentialAction/FitCompoundPotentialAction.hpp"

#include "CodePatterns/Log.hpp"

#include "Fragmentation/Homology/HomologyContainer.hpp"
#include "Fragmentation/Homology/HomologyGraph.hpp"
#include "Fragmentation/Summation/SetValues/Fragment.hpp"
#include "Potentials/Exceptions.hpp"
#include "Potentials/PotentialDeserializer.hpp"
#include "Potentials/PotentialFactory.hpp"
#include "Potentials/PotentialRegistry.hpp"
#include "Potentials/PotentialTrainer.hpp"
#include "Potentials/SerializablePotential.hpp"
#include "World.hpp"

using namespace MoleCuilder;

// and construct the stuff
#include "FitCompoundPotentialAction.def"
#include "Action_impl_pre.hpp"
/** =========== define the function ====================== */

ActionState::ptr PotentialFitCompoundPotentialAction::performCall() {
  // fragment specifies the homology fragment to use
  SerializablePotential::ParticleTypes_t fragmentnumbers =
      PotentialTrainer::getNumbersFromElements(params.fragment.get());

  // either charges and a potential is specified or a file
  if (boost::filesystem::exists(params.potential_file.get())) {
    std::ifstream returnstream(params.potential_file.get().string().c_str());
    if (returnstream.good()) {
      try {
        PotentialDeserializer deserialize(returnstream);
        deserialize();
      } catch (SerializablePotentialMissingValueException &e) {
        if (const std::string *key = boost::get_error_info<SerializablePotentialKey>(e))
          STATUS("Missing value when parsing information for potential "+*key+".");
        else
          STATUS("Missing value parsing information for potential with unknown key.");
        return Action::failure;
      } catch (SerializablePotentialIllegalKeyException &e) {
        if (const std::string *key = boost::get_error_info<SerializablePotentialKey>(e))
          STATUS("Illegal key parsing information for potential "+*key+".");
        else
          STATUS("Illegal key parsing information for potential with unknown key.");
        return Action::failure;
      }
    } else {
      STATUS("Failed to parse from "+params.potential_file.get().string()+".");
      return Action::failure;
    }
    returnstream.close();

    LOG(0, "STATUS: I'm training now a set of potentials parsed from "
        << params.potential_file.get().string() << " on a fragment "
        << fragmentnumbers << " on data from World's homologies.");

  } else {
      STATUS("No existing potential file given!");
      return Action::failure;
  }

  // parse homologies into container
  HomologyContainer &homologies = World::getInstance().getHomologies();

  // first we try to look into the HomologyContainer
  LOG(1, "INFO: Listing all present homologies ...");
  for (HomologyContainer::container_t::const_iterator iter =
      homologies.begin(); iter != homologies.end(); ++iter) {
    LOG(1, "INFO: graph " << iter->first
        << " has Fragment " << iter->second.fragment
        << ", associated energy " << iter->second.energy
        << ", and sampled grid integral " << iter->second.charge_distribution.integral()
        << ".");
  }

  // then we ought to pick the right HomologyGraph ...
  const HomologyGraph graph =
      PotentialTrainer::getFirstGraphwithSpecifiedElements(homologies,fragmentnumbers);
  if (graph != HomologyGraph()) {
    LOG(1, "First representative graph containing fragment "
        << fragmentnumbers << " is " << graph << ".");
  } else {
    STATUS("Specific fragment "+toString(fragmentnumbers)+" not found in homologies!");
    return Action::failure;
  }

  // training
  PotentialTrainer trainer;
  const bool status = trainer(
      homologies,
      graph,
      params.training_file.get(),
      params.threshold.get(),
      params.best_of_howmany.get());
  if (!status) {
    STATUS("No required parameter derivatives for a box constraint minimization known.");
    return Action::failure;
  }

  // store the potentials
  if (!params.potential_file.get().string().empty()) {
    PotentialTrainer::storePotentials(params.potential_file.get());
  }

  return Action::success;
}

ActionState::ptr PotentialFitCompoundPotentialAction::performUndo(ActionState::ptr _state) {
  return Action::success;
}

ActionState::ptr PotentialFitCompoundPotentialAction::performRedo(ActionState::ptr _state){
  return Action::success;
}

bool PotentialFitCompoundPotentialAction::canUndo() {
  return false;
}

bool PotentialFitCompoundPotentialAction::shouldUndo() {
  return false;
}
/** =========== end of function ====================== */
